# -*- project-name: VASM -*-
package VASM::Catalog::Message;

use strict;
use warnings;
use VASM::Tree;
use XML::Parser;
use XML::Writer;
use Encode qw/from_to/;
use Encode::Alias;
use Carp;

our $VERSION = '2.0';

# Known aliases for Encode -- the value returned by langinfo(CODESET) in
# VASM::Resource::Catalog::Message do not always agree with the codesets and
# aliases defined by Encode itself. Presently the only one I know of needs
# aliasing is TIS-620 (standard 8-bit encoding for Thai)
my %aliases = (
  # "I'm ^^relish^^, motherf---er! Mustard... ZAP YO DUMB ASS!"
  'TIS-620' => 'tis620'
);

# Now define them
for my $alias (keys %aliases) {
  define_alias($alias, $aliases{$alias});
}

sub new {
  # The IO::Handle to parse
  my ($self, %args) = @_; my $instance = {};

  $instance->{catalog} = VASM::Tree->new; # The message catalog itself
  $instance->{identifiers} = VASM::Tree->new; # Quantity identifiers
  bless $instance, $self;

  # If a file handle was given to parse:
  defined $args{handle} and $instance->Parse($args{handle});

  # If UTF-8 is not desired for rendered messages, an encoding may also be
  # specified. If this field is left blank, the attribute will remain
  # undefined.
  $instance->{encoding} = $args{encoding};

  return $instance;
}

sub Store {
  # $args{qw/id quantity/} represents the original message in English and the
  # quantity expressed. $args{message} is the translated message bearing all
  # these other qualities.
  my ($self, %args) = @_;

  # All traits must be defined
  croak 'All arguments must be given' 
    unless (grep { defined $args{$_} } qw/id quantity message/) == 3;

  # The quantity must also be defined
  croak 'Quantity undefined' unless $self->definedQuantity($args{quantity});

  # It is now very simple to store the message in the catalog itself
  $self->{catalog}->Store(@args{qw/id quantity message/});
  
  return;
}

sub SetQuantity {
  my ($self, $quantity, $identifier) = @_;

  # All traits must be defined
  croak 'All arguments must be given' unless @_ == 3;

  # Put the new identifier in the identifiers tree
  $self->{identifiers}->Store($quantity, qr/$identifier/);

  return;
}

sub matchQuantity {
  my ($self, $quantity) = @_;

  croak 'Quantity argument must be given' unless defined $quantity;

  # Iterate over the patterns provided in the identifiers tree
  for my $key ($self->{identifiers}->Children) {
    return $key if $quantity =~ $self->{identifiers}->Retrieve($key);
  }

  return;
}

sub definedQuantity {
  my ($self, $name) = @_;

  croak 'Name argument must be given' unless defined $name;

  return 1 if defined $self->{identifiers}->Retrieve($name);
  return;
}

sub Render {
  my $self = shift; my %args;

  # If there is a single argument, it is shorthand for specifying the id
  # attribute. Otherwise, the arguments specify a whole hash constructor.
  if (@_ == 1) { $args{id} = $_[0] }
  else { %args = @_ }
  
  # This will become a /linguistic/ quantity, as opposed to $args{quantity},
  # which is a number.
  my ($quantity, @arguments);
  
  # If no /numerical/ quantity was given, assume -1
  $args{quantity} = -1 unless defined $args{quantity};

  # At this point, quantity and id need to be defined in %args
  croak 'All arguments must be given' 
    unless (grep { defined $args{$_} } qw/quantity id/) == 2;
  
  # Here, give a linguistic identification to the raw number, and croak if
  # unsuccessful
  croak 'Quantity undefined'
    unless $quantity = $self->matchQuantity($args{quantity});

  # Return the formatted message with all remaining arguments, after removing
  # the extraneous whitespace from it
  my $message = $self->{catalog}->Retrieve($args{id}, $quantity);
  croak 'Message does not exist' unless defined $message;
  
  $message =~ s!(?:^\s*|\s*$)!!gs; $message =~ s!\s*\n\s*! !g;
  defined $args{arguments} and @arguments = @{ $args{arguments} };
  my $renderedMessage = sprintf $message, @arguments;

  # If an alternative encoding was given, recode to it before returning the
  # string
  from_to($renderedMessage, 'utf8', $self->{encoding})
    if defined $self->{encoding};

  return $renderedMessage;
}

# List the messages in a given catalog
sub List {
  my ($self) = @_;
  
  return $self->{catalog}->Children;
}

sub parseQuantity {
  my ($self, $attrs) = @_;
  # Quantity tags must have the 'match' and 'degree' attributes. 'match',
  # however, may be null.
  croak 'Attributes not given in <quantity> tag'
    unless exists $attrs->{match} and length $attrs->{degree};
  # Add the new quantity definition
  $self->SetQuantity($attrs->{degree}, $attrs->{match});  
}

sub parseMessage {
  my ($self, $messages) = @_;

  # Message tags must have the 'id' attribute
  croak '<message> tag lacks id attribute' unless length $messages->[0]->{id};

  # The remaining tags represent quantities and their translated messages
  for my $index (grep { $_ % 2 } (0..$#{ $messages })) {
    next if $messages->[$index] eq '0'; # Skip text elements

    # The quantity must be defined and the textual content of the tag must
    # also have a length > 0
    croak 'Undefined quantity, invalid or blank content in <message> tag'
      unless $messages->[$index + 1]->[1] eq '0'
        and length $messages->[$index + 1]->[2]
          and $self->definedQuantity($messages->[$index]);
    # And intern it...
    $self->Store(
      id => $messages->[0]->{id},
      quantity => $messages->[$index],
      message => $messages->[$index + 1]->[2]
    );
  }
}

# Piggyback on the facilities of XML::Parser to instantiate a message catalog
# from the contents of an XML file for which an IO::File is given as input.
sub Parse {
  my ($self, $handle) = @_;
  # If this fails, XML::Parser will emit an error message and end the program.
  # Unfortunately, this message is always English. :( There is at least an
  # error context, which can be universally understood by the programmer.
  my $parser = XML::Parser->new(Style => 'Tree', ErrorContext => 3);
  my $tree = $parser->parse($handle); my $tags = $tree->[1];
  
  # Main loop: iterate through tags under <catalog>
  for my $index (grep { $_ % 2 } (0..$#{ $tags })) {
    next if $tags->[$index] eq '0'; # Skip text elements

    if ($tags->[$index] eq 'quantity') {
      $self->parseQuantity($tags->[$index + 1]->[0]);
    } elsif ($tags->[$index] eq 'message') {
      $self->parseMessage($tags->[$index + 1]);
    } else {
      croak 'Invalid tag in message catalog';
    }
  }
  
  return;
}

sub Write {
  my ($self, $file) = @_;

  croak 'File argument must be given' unless defined $file;

  # Now create an XML::Writer instance and turn that puppy out
  my $xmlFile = XML::Writer->new(OUTPUT => $file, ENCODING => 'utf-8',
                                 DATA_MODE => 1, DATA_INDENT => 2);

  # Open up the document
  $xmlFile->xmlDecl;
  $xmlFile->startTag('catalog');
  $xmlFile->comment('Generated by VASM');

  # Emit the 'quantity' tags
  for my $quantity ($self->{identifiers}->Children) {
    $xmlFile->emptyTag(
      'quantity',
      degree => $quantity,
      match => $self->{identifiers}->Retrieve($quantity)
    );
  }

  # Now emit the 'message' tags
  for my $message ($self->List) {
    $xmlFile->startTag('message', id => $message);
    # Message and quantity
    for my $quantity ($self->{catalog}->Children($message)) {
      $xmlFile->dataElement(
        $quantity, $self->{catalog}->Retrieve($message, $quantity)
      );
    }
    $xmlFile->endTag('message');
  }

  $xmlFile->endTag('catalog'); $xmlFile->end; # End the document...

  return;
}

1;

__END__

=head1 NAME

VASM::Catalog::Message - i18n message catalogs with XML serialization

=head1 SYNOPSIS

    use VASM::Catalog::Message;

    my $instance = VASM::Catalog::Message->new($ioHandle);
    print $instance->Render(id => 'Phone Call', quantity => 0);
    # -> Niemand ruft mich an
    print $instance->Render(id => 'Phone Call', quantity => 1, 
                            arguments => [ 'Peter' ]);
    # -> Peter ruft mich an
    print $instance->Render(id => 'Phone Call', quantity => 2,
                            arguments => [ 'Peter und Sabine' ]);
    # -> Peter und Sabine rufen mich an

=head1 DESCRIPTION

VASM::Catalog::Message provides for message translation from a source language
(presumably English) to the user's native language, including flexible support
for quantities expressed in the language via regexes and arguments in the
traditional sprintf notation. In addition to the simple object interface,
VASM::Catalog::Message instances can digest an XML file containing quantity
definitions and translated messages. VASM::Catalog::Message even allows the
use of proper inflection or plural forms when issuing messages in the source
language itself, when varying quantities are expected.

=head1 METHODS

These are all the methods generally useful to a programmer utilizing the
VASM::Catalog::Message class. A few methods, such as matchQuantity, are really
only useful internally. If you wish to fully understand the inner workings of
the class, it will be necessary to peruse the source.

=over

=item new([ $handle ])

new constructs a new instance of the VASM::Catalog::Message class, and accepts
an optional argument, indicating an IO::Handle representing an XML file, whose
catalog definition you wish to introduce to the instance upon construction.
See L</FORMAT> below.

=item Store(id => ..., quantity => ..., message => ...)

The method Store accepts a hash constructor of three key/value pairs: 'id',
the untranslated source language message; 'quantity', a name designating a
linguistic quantity such as 'nullar', 'singular', 'paucal', or 'plural'; and
'message', the translated message itself, representing these other two
qualities. This translated message may contain all the directives accepted by
Perl's sprintf function; see L<perlfunc> and the Render method below. This
method returns nothing.

=item SetQuantity($quantity => $regex)

The SetQuantity method associates a named, gramatically significant quantity
with a regular expression, so that the use of a raw number in Render (see
below) will identify that proper gramatical form in emitting the translated
message; the names you give quantities are not important so long as the use is
consistent. Furthermore, the matching regular expression is usually anchored
on either side, to prevent a value like '10' to be considered nullar or
singular, for instance.

This method has a particularly interesting property: when multiple quantity
definitions might match a given number, the one interned in the instance
earlier takes precedence. Most languages, for instance, have some way of
expressing the idea of nothing (nullar), the idea of a single entity
(singular), and the idea of two or more entities (plural), some more elaborate
than others. Languages like Lihir and Polish really go to town here, but for a
relatively simple language like English, it's simple enough to add patterns
for the nullar and singular quantities and let a null pattern trap everything
else for plural:

    $instance->SetQuantity(nullar => '^0$');
    $instance->SetQuantity(singular => '^1$');
    $instance->SetQuantity(plural => '');

...and the XML catalog will let you do this, too. This implicit matching is
way more convenient than a fully qualified expression like
'^[2-9]|\p{IsDigit}{2,}$', which explictly matches plural quantities.

In cases where there is really no discernable quantity to speak of, I have
chosen the name 'sunya' and paired it with '^-1$'. 'Sunya' is the Sanskrit for
'void' or 'zero'; Thai/Lao speakers will know 'sunna', although 'sunya' itself
means 'promise' in those languages. That's awfully wry.

SetQuantity returns nothing. In that respect, its relation to its arguments
and return value is like mine to library books and the library itself.

=item Render(id => ..., [ quantity => .., arguments => ... ])

The method Render accepts a hash constructor similar to that prescribed for
the Store and SetQuantity methods (noticing a pattern?), whose keys and values
are id, again referring to the original source message; quantity, an integer
which introduces a grammatical context to the translated message and defaults
to -1 if not given (see SetQuantity above); and arguments, an optional listref
value containing values for any sprintf directives in the original message.
These arguments, if any, are applied to the translated message, which Render
returns. This, for example:

    $instance->Render(id => 'Greetings', arguments => [ qw/Hanumizzle/ ]);

...would yield 'Namaste, Hanumizzle', if the undefined or sunya value for the
message 'Greetings' was 'Namaste, %s'. If the target language(s) gramatically
inflected the contents of the strings given as arguments, it would be
necessary to compose calls to the Render object; that is, pass the rendition
of a string from a completely different catalog in as a member of the
arguments arrayref.

Render will also accept a single scalar argument, which will be construed as
the 'id' key in the arguments hash. quantity, as usual, will default to -1.
This is useful for the common idiom of an constant string such as 'OK' or
'Cancel', e.g., $instance->Render('OK');

=item List

Returns a list of the messages translated in the catalog. Each member of this
list will correspond to the 'id' field passed to the Store method.

=item Parse($handle)

This method accepts a single argument, an IO::Handle whose referent XML file
to intern in the instance's message collection. (See </FORMAT> below.) Parse
returns nothing, but will croak if unsuccessful.

=item Write($handle)

The Write method serves as the counterpart to the Parse method, accepting an
IO::Handle with which to serialize the invocant instance. The output produced
takes human beings into account, and will follow reasonable indenting
guidelines.

=back

=head1 FORMAT

As said before, VASM::Catalog::Message introduces an extremely simple XML
format for serializing message catalogs. It is almost self-explanatory upon
examination:

    <?xml version="1.0" encoding="utf-8"?>

    <catalog>
      <quantity degree='nullar' match='^0$'/>
      <quantity degree='singular' match='^1$'/>
      <quantity degree='plural' match=''/>
      <message id='Phone Call'>
        <nullar>Niemand ruft mich an</nullar>
        <singular>%s ruft mich an</singular>
        <plural>%s rufen mich an</plural>
      </message>
    </catalog>

'quantity' tags contain a 'degree' and 'match' attribute, which indicate a
linguistic quantity and matching regular expression, respectively. The 'match'
attribute may be empty, but 'degree' attributes may not; see rationale above.
Furthermore, any textual content of 'quantity' tags will simply be ignored, as
well as any other deviation that does not seriously conflict with the
semantics of the format itself. (To test this, try swapping the root tag name,
'catalog', with 'bananas'.) 'message' tags must have an 'id' attribute,
indicating the original message, and the remaining tags therein will be
interpreted as message translations for a given quantity, the only stipulation
being that they must correspond to a previously defined 'quantity'.

=head1 AUTHORS

hanumizzle L<mailto:hanumizzle@gmail.com>: principal author
cintyram: suggestion of better description of the package
YaP: the recoding facility and implementation details

=cut
